#
# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import datetime
import functools
import json
import pathlib

import pytest
import responses
import zeep.transports

import eopayment
from eopayment.common import EXPIRED, WAITING


@pytest.fixture
def saga(record_http_session):
    if type(record_http_session).__name__ != 'Mocker':
        from eopayment import saga

        saga._zeep_transport = zeep.transports.Transport(session=record_http_session)
        try:
            yield None
        finally:
            saga._zeep_transport = None
    else:
        yield


def factory(**kwargs):
    return eopayment.Payment(
        'saga',
        dict(
            {
                'num_service': '868',
                'compte': '70688',
                'automatic_return_url': 'https://automatic.notif.url/automatic/',
                'normal_return_url': 'https://normal.notif.url/normal/',
            },
            **kwargs,
        ),
    )


@pytest.fixture
def backend_factory(saga, target_url):
    return functools.partial(factory, base_url=target_url)


def test_features():
    payment = factory()
    assert payment.is_email_required


def test_error_parametrage(backend_factory):
    payment = backend_factory(num_service='1', compte='1')
    with pytest.raises(eopayment.PaymentException, match='Impossible de déterminer le paramétrage'):
        transaction_id, kind, url = payment.request(
            amount='10.00', email='john.doe@example.com', subject='Réservation concert XYZ numéro 1234'
        )


def test_error_no_email(backend_factory):
    payment = backend_factory()
    with pytest.raises(eopayment.PaymentException, match='.*not a valid email'):
        transaction_id, kind, url = payment.request(
            amount='10.00', email='', subject='Réservation concert XYZ numéro 1234'
        )


def test_request(backend_factory):
    transaction_id, kind, url = backend_factory().request(
        amount='10.00', email='john.doe@example.com', subject='Réservation concert XYZ numéro 1234'
    )
    assert transaction_id == '347b2060-1a37-11eb-af92-0213ad91a103'
    assert kind == eopayment.URL
    assert (
        url == 'https://www.tipi.budget.gouv.fr/tpa/paiementws.web?idop=347b2060-1a37-11eb-af92-0213ad91a103'
    )


def test_request_custom_num_service_and_compte(backend_factory):
    transaction_id, kind, url = backend_factory().request(
        amount='10.00',
        email='john.doe@example.com',
        subject='Réservation concert XYZ numéro 1234',
        num_service='1234',
        compte='abcd',
    )
    assert transaction_id == '347b2060-1a37-11eb-af92-0213ad91a103'
    assert kind == eopayment.URL
    assert (
        url == 'https://www.tipi.budget.gouv.fr/tpa/paiementws.web?idop=347b2060-1a37-11eb-af92-0213ad91a103'
    )


def test_response(backend_factory):
    response = backend_factory().response('idop=28b52f40-1ace-11eb-8ce3-0213ad91a104', redirect=False)
    assert json.dumps(response.as_dict())
    assert response.__dict__ == {
        'bank_data': {
            'email': 'john.doe@entrouvert.com',
            'etat': 'paye',
            'id_tiers': '-1',
            'montant': '10.00',
            'num_service': '868',
            'numcp': '70688',
            'numcpt_lib_ecriture': 'COUCOU',
        },
        'bank_status': 'paid',
        'order_id': '28b52f40-1ace-11eb-8ce3-0213ad91a104',
        'result': 3,
        'return_content': None,
        'signed': True,
        'test': False,
        'transaction_date': None,
        'transaction_id': '28b52f40-1ace-11eb-8ce3-0213ad91a104',
    }
    # Check bank_data is JSON serializable
    json.dumps(response.bank_data)


def test_payment_status(backend_factory):
    response = backend_factory().payment_status('28b52f40-1ace-11eb-8ce3-0213ad91a104')
    assert response.__dict__ == {
        'bank_data': {
            'email': 'john.doe@entrouvert.com',
            'etat': 'paye',
            'id_tiers': '-1',
            'montant': '10.00',
            'num_service': '868',
            'numcp': '70688',
            'numcpt_lib_ecriture': 'COUCOU',
        },
        'bank_status': 'paid',
        'order_id': '28b52f40-1ace-11eb-8ce3-0213ad91a104',
        'result': 3,
        'return_content': None,
        'signed': True,
        'test': False,
        'transaction_date': None,
        'transaction_id': '28b52f40-1ace-11eb-8ce3-0213ad91a104',
    }
    # Check bank_data is JSON serializable
    json.dumps(response.bank_data)


def test_payment_status_cancelled(backend_factory):
    response = backend_factory().payment_status('28b52f40-1ace-11eb-8ce3-0213ad91a104')
    assert response.__dict__ == {
        'bank_data': {
            'email': 'john.doe@entrouvert.com',
            'etat': 'abandon',
            'id_tiers': '-1',
            'montant': '10.00',
            'num_service': '868',
            'numcp': '70688',
            'numcpt_lib_ecriture': 'COUCOU',
        },
        'bank_status': 'cancelled',
        'order_id': '28b52f40-1ace-11eb-8ce3-0213ad91a104',
        'result': 5,
        'return_content': None,
        'signed': True,
        'test': False,
        'transaction_date': None,
        'transaction_id': '28b52f40-1ace-11eb-8ce3-0213ad91a104',
    }
    # Check bank_data is JSON serializable
    json.dumps(response.bank_data)


def test_payment_status_error(backend_factory):
    with pytest.raises(eopayment.PaymentException, match=r'Saga error message: foobar'):
        backend_factory().payment_status('28b52f40-1ace-11eb-8ce3-0213ad91a104')


def test_username_token_auth(record_http_session, backend_factory):
    backend = backend_factory(login='foo', password='bar')
    backend.request(
        amount='10.00',
        email='john.doe@example.com',
        subject='Réservation concert XYZ numéro 1234',
    )
    assert (
        record_http_session.requests[0].url == 'https://target.url/paiement_internet_ws_ministere_secure?wsdl'
    )
    assert record_http_session.requests[1].url == 'https://target.url/paiement_internet_ws_ministere_secure'


def test_payment_status_waiting_P5(backend_factory):
    transaction_date = datetime.datetime.now() - datetime.timedelta(minutes=19)
    assert (
        backend_factory()
        .payment_status(
            transaction_id='28b52f40-1ace-11eb-8ce3-0213ad91a104', transaction_date=transaction_date
        )
        .result
        == WAITING
    )


def test_payment_status_expired_P5(backend_factory):
    transaction_date = datetime.datetime.now() - datetime.timedelta(minutes=21)
    assert (
        backend_factory()
        .payment_status(
            transaction_id='28b52f40-1ace-11eb-8ce3-0213ad91a104', transaction_date=transaction_date
        )
        .result
        == EXPIRED
    )


def test_payment_status_expired_P1(backend_factory):
    assert backend_factory().payment_status('28b52f40-1ace-11eb-8ce3-0213ad91a104').result == EXPIRED


@responses.activate
def test_timeout(target_url):
    responses.get(
        'https://target.url/paiement_internet_ws_ministere?wsdl',
        body=pathlib.Path('tests/data/saga_paiement_internet_ws_ministere.wsdl').read_text(),
    )

    soap_response = json.loads(pathlib.Path('tests/data/saga-test_request.json').read_text())[1][1]

    post_mock = responses.post(
        'https://saga.example.com/saga/paiement_internet_ws_ministere', body=soap_response
    )

    factory(base_url=target_url).request(
        amount='10.00',
        email='john.doe@example.com',
        subject='Réservation concert XYZ numéro 1234',
        timeout=3,
    )

    assert post_mock.calls[0].request.req_kwargs['timeout'] == 15



# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import copy
import datetime
import decimal
import functools
import os
import random
import re
import unicodedata
import xml.etree.ElementTree as ET
import zoneinfo
from urllib.parse import parse_qs

import requests
import zeep
import zeep.exceptions

from .common import (
    CANCELLED,
    DENIED,
    ERROR,
    EXPIRED,
    PAID,
    URL,
    WAITING,
    PaymentCommon,
    PaymentException,
    PaymentResponse,
    ResponseError,
    _,
    force_text,
)
from .systempayv2 import isonow

# The URL of the WSDL published in the documentation is still wrong, it
# references XSD files which are not resolvable :/ we must use this other URL
# where the XSD files are resolvable. To not depend too much on those files, we
# provide copy in eopayment. The PayFiP development team is full of morons.
WSDL_URL = 'https://www.payfip.gouv.fr/tpa/services/securite?wsdl'  # noqa: E501
PAYMENT_URL = 'https://www.payfip.gouv.fr/tpa/paiementws.web'

REFDET_RE = re.compile(r'^[A-Za-z0-9]{1,30}$')


def clear_namespace(element):
    def helper(element):
        if element.tag.startswith('{'):
            element.tag = element.tag[element.tag.index('}') + 1 :]
        for subelement in element:
            helper(subelement)

    element = copy.deepcopy(element)
    helper(element)
    return element


def normalize_objet(objet):
    '''Make objet a string of 100 chars in alphabet [A-Za-z0-9 ]'''
    if not objet:
        return objet

    objet = force_text(objet)
    objet = unicodedata.normalize('NFKD', objet).encode('ascii', 'ignore').decode()
    objet = re.sub(r'[\'-]', ' ', objet).strip()
    objet = re.sub(r'[^A-Za-z0-9 ]', '', objet).strip()
    objet = re.sub(r'[\s]+', ' ', objet)
    return objet[:100]


class PayFiPError(PaymentException):
    def __init__(self, code, message, origin=None):
        self.code = code
        self.message = message
        self.origin = origin
        args = [code, message]
        if origin:
            args.append(origin)
        super().__init__(*args)


class PayFiP:
    '''Encapsulate SOAP web-services of PayFiP'''

    default_timeout = 15

    def __init__(self, wsdl_url=None, service_url=None, zeep_client_kwargs=None, use_local_wsdl=True):
        # use cached WSDL
        if (not wsdl_url or (wsdl_url == WSDL_URL)) and use_local_wsdl:
            base_path = os.path.join(os.path.dirname(__file__), 'resource', 'PaiementSecuriseService.wsdl')
            wsdl_url = 'file://%s' % base_path
        self.wsdl_url = wsdl_url
        self.service_url = service_url
        self.zeep_client_kwargs = zeep_client_kwargs

    @property
    def client(self):
        if not hasattr(self, '_client'):
            try:
                self._client = zeep.Client(self.wsdl_url or WSDL_URL, **(self.zeep_client_kwargs or {}))
            except Exception as e:
                raise PayFiPError('Cound not initialize the SOAP client', e)
        return self._client

    def fault_to_exception(self, fault):
        if (
            fault.message != 'fr.gouv.finances.cp.tpa.webservice.exceptions.FonctionnelleErreur'
            or fault.detail is None
        ):
            return
        detail = clear_namespace(fault.detail)
        code = detail.find('FonctionnelleErreur/code')
        if code is None or not code.text:
            return PayFiPError('inconnu', ET.tostring(detail))
        descriptif = detail.find('FonctionnelleErreur/descriptif')
        libelle = detail.find('FonctionnelleErreur/libelle')
        return PayFiPError(
            code=code.text,
            message=(descriptif is not None and descriptif.text)
            or (libelle is not None and libelle.text)
            or '',
        )

    def _perform(self, request_qname, operation, timeout=None, **kwargs):
        RequestType = self.client.get_type(request_qname)  # noqa: E501
        try:
            with self.client.transport.settings(timeout=timeout or self.default_timeout):
                return getattr(self.client.service, operation)(RequestType(**kwargs))
        except zeep.exceptions.Fault as fault:
            raise self.fault_to_exception(fault) or PayFiPError('unknown', fault.message, fault)
        except zeep.exceptions.Error as zeep_error:
            raise PayFiPError('SOAP error', str(zeep_error), zeep_error)
        except requests.RequestException as e:
            raise PayFiPError('HTTP error', e)
        except Exception as e:
            raise PayFiPError('Unexpected error', e)

    def get_info_client(self, numcli, timeout=None):
        return self._perform(
            '{http://securite.service.tpa.cp.finances.gouv.fr/reponse}RecupererDetailClientRequest',
            'recupererDetailClient',
            timeout=timeout,
            numCli=numcli,
        )

    def get_idop(
        self,
        numcli,
        saisie,
        exer,
        refdet,
        montant,
        mel,
        url_notification,
        url_redirect,
        objet=None,
        timeout=None,
    ):
        return self._perform(
            '{http://securite.service.tpa.cp.finances.gouv.fr/requete}CreerPaiementSecuriseRequest',
            'creerPaiementSecurise',
            numcli=numcli,
            saisie=saisie,
            exer=exer,
            montant=montant,
            refdet=refdet,
            mel=mel,
            urlnotif=url_notification,
            urlredirect=url_redirect,
            objet=objet,
            timeout=timeout,
        )

    def get_info_paiement(self, idop, timeout=None):
        return self._perform(
            '{http://securite.service.tpa.cp.finances.gouv.fr/reponse}RecupererDetailPaiementSecuriseRequest',
            'recupererDetailPaiementSecurise',
            idOp=idop,
            timeout=timeout,
        )


class Payment(PaymentCommon):
    """Produce requests for and verify response from the TIPI online payment
    processor from the French Finance Ministry.

    """

    description = {
        'caption': 'TIPI, Titres Payables par Internet',
        'parameters': [
            {
                'name': 'numcli',
                'caption': _('Client number'),
                'help_text': _('6 digits number provided by DGFIP'),
                'validation': lambda s: str.isdigit(s) and len(s) == 6,
                'required': True,
            },
            {
                'name': 'saisie',
                'caption': _('Payment type'),
                'default': 'T',
                'choices': [
                    ('T', _('test')),
                    ('X', _('activation')),
                    ('W', _('production')),
                ],
            },
            {
                'name': 'normal_return_url',
                'caption': _('User return URL'),
                'required': True,
            },
            {
                'name': 'automatic_return_url',
                'caption': _('Asynchronous return URL'),
                'required': True,
            },
        ],
    }

    min_time_between_transactions = 60 * 20  # 20 minutes

    minimal_amount = decimal.Decimal('1.0')
    maximal_amount = decimal.Decimal('100000.0')

    email_is_required = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.payfip = PayFiP()

    def _generate_refdet(self):
        return '%s%010d' % (isonow(), random.randint(1, 1000000000))

    def request(
        self,
        amount,
        email,
        refdet=None,
        exer=None,
        orderid=None,
        subject=None,
        transaction_id=None,
        timeout=None,
        **kwargs,
    ):
        montant = self.clean_amount(amount, max_amount=100000)

        numcli = self.numcli
        urlnotif = self.automatic_return_url
        urlredirect = self.normal_return_url

        if not exer:
            exer = str(datetime.date.today().year)

        if refdet:
            pass
        elif transaction_id and REFDET_RE.match(transaction_id):
            refdet = transaction_id
        elif orderid and REFDET_RE.match(orderid):
            refdet = orderid
        else:
            refdet = self._generate_refdet()

        objet_parts = []
        if orderid and refdet != orderid:
            objet_parts.extend(['O', orderid])
        if subject:
            if objet_parts:
                objet_parts.append('S')
            objet_parts.append(subject)
        if transaction_id and refdet != transaction_id:
            objet_parts.extend(['T', transaction_id])
        objet = normalize_objet(' '.join(objet_parts))

        mel = email
        if hasattr(mel, 'decode'):
            mel = email.decode('ascii')
        try:
            if '@' not in mel:
                raise ValueError('no @ in MEL')
            if not (6 <= len(mel) <= 80):
                raise ValueError('len(MEL) is invalid, must be between 6 and 80')
        except Exception as e:
            raise ValueError('MEL is not a valid email, %r' % mel, e)

        # check saisie
        saisie = self.saisie
        if saisie not in ('T', 'X', 'W'):
            raise ValueError('SAISIE invalid format, %r, must be M, T, X or A' % saisie)

        idop = self.payfip.get_idop(
            numcli=numcli,
            saisie=saisie,
            exer=exer,
            refdet=refdet,
            montant=montant,
            mel=mel,
            objet=objet or None,
            url_notification=urlnotif,
            url_redirect=urlredirect,
            timeout=timeout,
        )

        return str(idop), URL, PAYMENT_URL + '?idop=%s' % idop

    P5_WAITING_IN_MINUTES = 120  # wait at least 2 hours
    CANCELLED_WAITING_IN_MINUTES = 2

    def payment_status(self, transaction_id, transaction_date=None, timeout=None, **kwargs):
        # idop are valid for 15 minutes after their generation
        # between generation and payment, any call to get_info_paiement() will return a PayFiPError with code=P5
        # before the end of the 15 minutes it can mean the payment is in progress
        # after the 15 minutes period it means the payment will never happen,
        # and after one day the code will change for P1, meaning the idop is
        # now unknown as it as been cleaned by the night cleaning job.
        #
        # So in order to interpret the meaning of PayFiP error codes we need
        # the date of the start of the transaction and add to it some margin
        # to.
        idop = transaction_id
        if transaction_date:
            if transaction_date.tzinfo:  # date is aware
                now = datetime.datetime.now(tz=datetime.timezone.utc)
            else:
                now = datetime.datetime.now()
            delta = now - transaction_date
        else:
            delta = datetime.timedelta(seconds=0)

        # set the threshold between transaction 'in progress' and 'expired' at self.P5_WAITING_IN_MINUTES minutes
        threshold = datetime.timedelta(minutes=self.P5_WAITING_IN_MINUTES)

        # for cancelled state, wait for self.CANCELLED_WAITING_IN_MINUTES minutes
        cancelled_threshold = datetime.timedelta(minutes=self.CANCELLED_WAITING_IN_MINUTES)

        try:
            response = self.payfip.get_info_paiement(idop, timeout=timeout)
        except PayFiPError as e:
            if e.code == 'P1' or (e.code == 'P5' and delta >= threshold):
                return PaymentResponse(result=EXPIRED, signed=True, order_id=transaction_id)
            if e.code == 'P5' and delta < threshold:
                return PaymentResponse(result=WAITING, signed=True, order_id=transaction_id)
            raise e
        payfip_transaction_datetime = self.get_transaction_datetime(response)

        if payfip_transaction_datetime is not None:
            delta = datetime.datetime.now(tz=datetime.timezone.utc) - payfip_transaction_datetime

        eopayment_response = self.payfip_response_to_eopayment_response(idop, response)
        # convert CANCELLED to WAITING during the first 20 minutes
        if eopayment_response.result == CANCELLED and delta < cancelled_threshold:
            eopayment_response.result = WAITING
            eopayment_response.bank_status = (
                '%s - still waiting as idop is still active' % eopayment_response.bank_status
            )
        return eopayment_response

    def response(self, query_string, **kwargs):
        fields = parse_qs(query_string, True)
        idop = (fields.get('idop') or [None])[0]

        if not idop:
            raise ResponseError('missing idop parameter in query string')

        return self.payment_status(idop)

    @classmethod
    def get_transaction_datetime(self, response):
        dattrans = response.dattrans
        heurtrans = response.heurtrans
        if not dattrans:
            return None
        if not heurtrans:
            return None
        full_string = f'{dattrans}{heurtrans}'
        try:
            dt = datetime.datetime.strptime(full_string, '%d%m%Y%H%M')
        except ValueError as e:
            return None
        localized_dt = dt.replace(tzinfo=zoneinfo.ZoneInfo('Europe/Paris'))
        return localized_dt

    @classmethod
    def payfip_response_to_eopayment_response(cls, idop, response):
        if response.resultrans == 'P':
            result = PAID
            bank_status = 'paid CB'
        elif response.resultrans == 'V':
            result = PAID
            bank_status = 'paid direct debit'
        elif response.resultrans == 'R':
            result = DENIED
            bank_status = 'refused CB'
        elif response.resultrans == 'Z':
            result = DENIED
            bank_status = 'refused direct debit'
        elif response.resultrans == 'A':
            result = CANCELLED
            bank_status = 'cancelled CB'
        else:
            result = ERROR
            bank_status = 'unknown result code: %r' % response.resultrans

        transaction_id = response.refdet
        transaction_id += ' ' + idop
        if response.numauto:
            transaction_id += ' ' + response.numauto

        return PaymentResponse(
            result=result,
            bank_status=bank_status,
            signed=True,
            bank_data={k: response[k] for k in response},
            order_id=idop,
            transaction_id=transaction_id,
            test=response.saisie == 'T',
        )

    @classmethod
    def guess(self, *, method=None, query_string=None, body=None, headers=None, backends=(), **kwargs):
        for content in [query_string, body]:
            if isinstance(content, bytes):
                try:
                    content = content.decode()
                except UnicodeDecodeError:
                    pass
            if isinstance(content, str):
                fields = parse_qs(content)
                if set(fields) == {'idOp'}:
                    return fields['idOp'][0]
        return None


if __name__ == '__main__':
    import click

    def show_payfip_error(func):
        @functools.wraps(func)
        def f(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except PayFiPError as e:
                click.echo(click.style('PayFiP ERROR : %s "%s"' % (e.code, e.message), fg='red'))

        return f

    @click.group()
    @click.option('--wsdl-url', default=None)
    @click.option('--service-url', default=None)
    @click.pass_context
    def main(ctx, wsdl_url, service_url):
        import logging

        logging.basicConfig(level=logging.INFO)
        # hide warning from zeep
        logging.getLogger('zeep.wsdl.bindings.soap').level = logging.ERROR

        ctx.obj = PayFiP(wsdl_url=wsdl_url, service_url=service_url)

    def numcli(ctx, param, value):
        if not isinstance(value, str) or len(value) != 6 or not value.isdigit():
            raise click.BadParameter('numcli must a 6 digits number')
        return value

    @main.command()
    @click.argument('numcli', callback=numcli, type=str)
    @click.pass_obj
    @show_payfip_error
    def info_client(payfip, numcli):
        response = payfip.get_info_client(numcli)
        for key in response:
            print('%15s:' % key, response[key])

    @main.command()
    @click.argument('numcli', callback=numcli, type=str)
    @click.option('--saisie', type=click.Choice(['T', 'X', 'W']), required=True)
    @click.option('--exer', type=str, required=True)
    @click.option('--montant', type=int, required=True)
    @click.option('--refdet', type=str, required=True)
    @click.option('--mel', type=str, required=True)
    @click.option('--url-notification', type=str, required=True)
    @click.option('--url-redirect', type=str, required=True)
    @click.option('--objet', default=None, type=str)
    @click.pass_obj
    @show_payfip_error
    def get_idop(payfip, numcli, saisie, exer, montant, refdet, mel, objet, url_notification, url_redirect):
        idop = payfip.get_idop(
            numcli=numcli,
            saisie=saisie,
            exer=exer,
            montant=montant,
            refdet=refdet,
            mel=mel,
            objet=objet,
            url_notification=url_notification,
            url_redirect=url_redirect,
        )
        print('idOp:', idop)
        print(PAYMENT_URL + '?idop=%s' % idop)

    @main.command()
    @click.argument('idop', type=str)
    @click.pass_obj
    @show_payfip_error
    def info_paiement(payfip, idop):
        print(payfip.get_info_paiement(idop))

    main()



# eopayment - online payment library

import contextlib
import json
import os

import pytest
from responses import RequestsMock, _recorder

from eopayment import Payment, common
from eopayment.worldline import Payment as WorldlinePayment
from eopayment.worldline import WorldlineAPI


@pytest.fixture
def worldline_params(backend_params):
    params = {
        'pspid': 'Entrouvert1',
        'api_key': '1234',
        'api_secret': '4567',
        **{key[10:]: value for key, value in backend_params.items() if key.startswith('worldline_')},
    }
    return params


@pytest.fixture
def api(worldline_params):
    return WorldlineAPI(
        url=WorldlinePayment.preprod_url,
        **worldline_params,
    )


HOSTEDCHECKOUTS_RESPONSE = {
    'RETURNMAC': 'ab874513-9f40-4ca0-a7c7-8dc14bc4e09d',
    'hostedCheckoutId': '4306873688',
    'merchantReference': '1234',
    'partialRedirectUrl': 'preprod.direct.worldline-solutions.com/hostedcheckout/PaymentMethods/Selection/ca5387a6432e47ed9e2ad17b27bf8385',
    'redirectUrl': 'https://payment.preprod.direct.worldline-solutions.com/hostedcheckout/PaymentMethods/Selection/ca5387a6432e47ed9e2ad17b27bf8385',
}


@pytest.fixture
def hostedcheckouts_mock(rsps):
    return rsps.post(
        'https://payment.preprod.direct.worldline-solutions.com/v2/Entrouvert1/hostedcheckouts',
        json=HOSTEDCHECKOUTS_RESPONSE,
    )


@pytest.fixture
def rsps_recorder(request):
    module_name = request.module.__name__.split('test_', 1)[-1]
    function_name = request.function.__name__
    save = request.config.getoption('--save-http-session')
    dir_path = f'tests/data/{module_name}/'
    given_name = getattr(request, 'param', None)
    path = f'{dir_path}{given_name or function_name}.yaml'

    os.makedirs(dir_path, exist_ok=True)

    with contextlib.ExitStack() as stack:
        if save:
            rsps = stack.enter_context(_recorder.Recorder())
        else:
            rsps = stack.enter_context(RequestsMock())
            rsps._add_from_file(path)
        yield rsps
        if save:
            print(f'Saving HTTP responses to {path}.')
            rsps.dump_to_file(path)


class TestWorldlineAPI:
    def test_hosted_checkout(self, api, rsps_recorder):
        assert api
        resp = api.hosted_checkout(
            amount=1000,
            return_url='https://example.com/return',
            reference='1234',
        )
        assert set(resp) == {
            'RETURNMAC',
            'hostedCheckoutId',
            'merchantReference',
            'partialRedirectUrl',
            'redirectUrl',
        }
        assert len(rsps_recorder.calls) == 1
        last_call = json.loads(rsps_recorder.calls[-1].request.body)
        assert last_call == {
            'cardPaymentMethodSpecificInput': {
                'authorizationMode': 'SALE',
            },
            'hostedCheckoutSpecificInput': {
                'returnUrl': 'https://example.com/return',
                'showResultPage': True,
                'allowedNumberOfPaymentAttempts': 1,
                'sessionTimeout': 300,
            },
            'order': {
                'amountOfMoney': {'amount': 1000, 'currencyCode': 'EUR'},
                'references': {'merchantReference': '1234'},
            },
        }

    @pytest.mark.parametrize('rsps_recorder', ['test_hosted_checkout'], indirect=True)
    def test_hosted_checkout_all_inputs(self, api, rsps_recorder):
        assert api
        resp = api.hosted_checkout(
            amount=1000,
            email='john.doe@example.com',
            phone='+33645454545',
            first_name='John',
            last_name='Doe',
            return_url='https://example.com/return',
            reference='1234',
        )
        assert set(resp) == {
            'RETURNMAC',
            'hostedCheckoutId',
            'merchantReference',
            'partialRedirectUrl',
            'redirectUrl',
        }
        assert len(rsps_recorder.calls) == 1
        last_call = json.loads(rsps_recorder.calls[-1].request.body)
        assert last_call == {
            'cardPaymentMethodSpecificInput': {
                'authorizationMode': 'SALE',
            },
            'hostedCheckoutSpecificInput': {
                'returnUrl': 'https://example.com/return',
                'showResultPage': True,
                'allowedNumberOfPaymentAttempts': 1,
                'sessionTimeout': 300,
            },
            'order': {
                'amountOfMoney': {'amount': 1000, 'currencyCode': 'EUR'},
                'customer': {
                    'contactDetails': {'emailAddress': 'john.doe@example.com', 'phoneNumber': '+33645454545'},
                    'personalInformation': {
                        'name': {
                            'firstName': 'John',
                            'surName': 'Doe',
                        },
                    },
                },
                'references': {'merchantReference': '1234'},
            },
        }

    STATUS_RESPONSE = {
        'createdPaymentOutput': {
            'payment': {
                'hostedCheckoutSpecificOutput': {
                    'hostedCheckoutId': '4311628069',
                },
                'id': '4311628069_0',
                'paymentOutput': {
                    'acquiredAmount': {
                        'amount': 10,
                        'currencyCode': 'EUR',
                    },
                    'amountOfMoney': {
                        'amount': 10,
                        'currencyCode': 'EUR',
                    },
                    'cardPaymentMethodSpecificOutput': {
                        'acquirerInformation': {
                            'name': 'ACQUIRER',
                        },
                        'authorisationCode': '837011',
                        'card': {
                            'bin': '51370098',
                            'cardNumber': '************3438',
                            'countryCode': 'BE',
                            'expiryDate': '0925',
                        },
                        'fraudResults': {
                            'avsResult': 'U',
                            'cvvResult': 'M',
                            'fraudServiceResult': 'no-advice',
                        },
                        'paymentAccountReference': 'FY2XWZZAYR8IEM8JSR47W5OUGBWVU',
                        'paymentProductId': 3,
                        'schemeReferenceData': '462482174457',
                        'threeDSecureResults': {
                            'acsTransactionId': '30085389-CFC5-46D1-AF3C-B5D8F7E486BC',
                            'authenticationStatus': 'Y',
                            'cavv': 'AAABBEg0VhI0VniQEjRWAAAAAAA=',
                            'dsTransactionId': 'f25084f0-5b16-4c0a-ae5d-b24808a95e4b',
                            'eci': '5',
                            'flow': 'frictionless',
                            'liability': 'issuer',
                            'schemeEci': '02',
                            'version': '2.2.0',
                            'xid': 'NDMxMTYyODA2OQ==',
                        },
                    },
                    'customer': {
                        'device': {
                            'ipAddressCountryCode': '99',
                        },
                    },
                    'paymentMethod': 'card',
                    'references': {
                        'merchantReference': '95ed629d7b484eaf91fd263498e15afe #1234',
                    },
                },
                'status': 'PENDING_CAPTURE',
                'statusOutput': {
                    'isAuthorized': True,
                    'isCancellable': True,
                    'isRefundable': False,
                    'statusCategory': 'PENDING_MERCHANT',
                    'statusCode': 5,
                },
            },
            'paymentStatusCategory': 'SUCCESSFUL',
        },
        'status': 'PAYMENT_CREATED',
    }

    def test_hosted_checkout_status(self, api, rsps_recorder):
        assert api.hosted_checkout_status(hosted_checkout_id='4315624881') == {
            'createdPaymentOutput': {
                'payment': {
                    'hostedCheckoutSpecificOutput': {
                        'hostedCheckoutId': '4315624881',
                    },
                    'id': '4315624881_0',
                    'paymentOutput': {
                        'acquiredAmount': {
                            'amount': 1010,
                            'currencyCode': 'EUR',
                        },
                        'amountOfMoney': {
                            'amount': 1010,
                            'currencyCode': 'EUR',
                        },
                        'cardPaymentMethodSpecificOutput': {
                            'acquirerInformation': {
                                'name': 'ACQUIRER',
                            },
                            'authorisationCode': '493667',
                            'card': {
                                'bin': '51370098',
                                'cardNumber': '************3438',
                                'countryCode': 'BE',
                                'expiryDate': '0129',
                            },
                            'fraudResults': {
                                'avsResult': 'U',
                                'cvvResult': 'M',
                                'fraudServiceResult': 'no-advice',
                            },
                            'paymentAccountReference': '3XIWXHQPYAZUF5CVKLQE2F7WFKIV7',
                            'paymentProductId': 3,
                            'schemeReferenceData': '151645221224',
                            'threeDSecureResults': {
                                'acsTransactionId': '30085389-CFC5-46D1-AF3C-B5D8F7E486BC',
                                'authenticationStatus': 'Y',
                                'cavv': 'AAABBEg0VhI0VniQEjRWAAAAAAA=',
                                'dsTransactionId': 'f25084f0-5b16-4c0a-ae5d-b24808a95e4b',
                                'eci': '5',
                                'flow': 'frictionless',
                                'liability': 'issuer',
                                'schemeEci': '02',
                                'version': '2.2.0',
                                'xid': 'NDMxNTYyNDg4MQ==',
                            },
                        },
                        'customer': {
                            'device': {
                                'ipAddressCountryCode': '99',
                            },
                        },
                        'paymentMethod': 'card',
                        'references': {
                            'merchantReference': 'abcd5678 #1234',
                        },
                    },
                    'status': 'CAPTURED',
                    'statusOutput': {
                        'isAuthorized': False,
                        'isCancellable': False,
                        'isRefundable': True,
                        'statusCategory': 'COMPLETED',
                        'statusCode': 9,
                    },
                },
                'paymentStatusCategory': 'SUCCESSFUL',
            },
            'status': 'PAYMENT_CREATED',
        }


@pytest.fixture
def payment(worldline_params):
    return Payment(
        'worldline',
        {
            'normal_return_url': 'https://example.com/return',
            'automatic_return_url': 'https://example.com/callback',
            'environment': 'preprod',
            **worldline_params,
        },
    )


class TestEopaymentAPI:
    def test_constructor(self, payment):
        assert payment
        params = {param['name']: param for param in payment.get_parameters()}
        assert set(params) == {
            'environment',
            'pspid',
            'api_key',
            'api_secret',
            'automatic_return_url',
            'normal_return_url',
        }
        assert set(params['environment']['choices']) == {'prod', 'preprod'}

    def test_direct_sale(self, payment, rsps_recorder):
        trans_id, _, _ = payment.request(
            amount='10.10',
            orderid='#1234',
            transaction_id='abcd5678',
            email='john.doe@example.com',
            first_name='John',
            last_name='Doe',
            phone='+33641414141',
        )

        if len(rsps_recorder.calls):
            last_call = json.loads(rsps_recorder.calls[-1].request.body)
            assert last_call == {
                'cardPaymentMethodSpecificInput': {
                    'authorizationMode': 'SALE',
                },
                'hostedCheckoutSpecificInput': {
                    'returnUrl': 'https://example.com/return',
                    'showResultPage': True,
                    'allowedNumberOfPaymentAttempts': 1,
                    'sessionTimeout': 300,
                },
                'order': {
                    'amountOfMoney': {'amount': 1010, 'currencyCode': 'EUR'},
                    'references': {'merchantReference': 'abcd5678 #1234'},
                    'customer': {
                        'contactDetails': {
                            'emailAddress': 'john.doe@example.com',
                            'phoneNumber': '+33641414141',
                        },
                        'personalInformation': {
                            'name': {
                                'firstName': 'John',
                                'surName': 'Doe',
                            },
                        },
                    },
                },
            }

        # wait for payment...
        query_string = f'RETURNMAC=dd70af3f-31ed-4366-a5c9-d8e2f8e0fa55&hostedCheckoutId={trans_id}'
        response = payment.response(query_string, redirect=True, order_id_hint='#1234')
        assert json.dumps(response.as_dict())
        assert response.test
        assert response.order_id == trans_id
        assert response.transaction_id == trans_id
        assert response.bank_status == 'CAPTURED'
        assert (
            response.bank_data['hosted_checkout']['createdPaymentOutput']['payment']['status'] == 'CAPTURED'
        )
        assert response.result == common.PAID

        # wait for capture...
        diagnostic_response = payment.diagnostic(bank_data=response.bank_data)
        assert diagnostic_response['payment_details']['status'] == 'CAPTURED'

    def test_authorization_capture(self, payment, rsps_recorder):
        trans_id, _, _ = payment.request(
            amount='10.10', orderid='#1234', transaction_id='abcd5678', authorization_mode='PRE_AUTHORIZATION'
        )
        if len(rsps_recorder.calls):
            last_call = json.loads(rsps_recorder.calls[-1].request.body)
            assert last_call == {
                'cardPaymentMethodSpecificInput': {
                    'authorizationMode': 'PRE_AUTHORIZATION',
                },
                'hostedCheckoutSpecificInput': {
                    'returnUrl': 'https://example.com/return',
                    'showResultPage': True,
                    'allowedNumberOfPaymentAttempts': 1,
                    'sessionTimeout': 300,
                },
                'order': {
                    'amountOfMoney': {'amount': 1010, 'currencyCode': 'EUR'},
                    'references': {'merchantReference': 'abcd5678 #1234'},
                },
            }

        # wait for payment...
        query_string = f'RETURNMAC=dd70af3f-31ed-4366-a5c9-d8e2f8e0fa55&hostedCheckoutId={trans_id}'
        response = payment.response(query_string, redirect=True, order_id_hint='#1234')
        assert response.test
        assert response.order_id == trans_id
        assert response.transaction_id == trans_id
        assert response.bank_status == 'PENDING_CAPTURE'
        assert (
            response.bank_data['hosted_checkout']['createdPaymentOutput']['payment']['status']
            == 'PENDING_CAPTURE'
        )
        assert response.result == common.PAID

        # capture the authorization, capture can be full or partial.
        validate_response = payment.validate('10.10', response.bank_data)
        assert (
            validate_response['hosted_checkout']['createdPaymentOutput']['payment']['status']
            == 'PENDING_CAPTURE'
        )
        assert validate_response['payment_details']['status'] == 'CAPTURE_REQUESTED'
        diagnostic_response = payment.diagnostic(bank_data=response.bank_data)
        assert diagnostic_response['payment_details']['status'] == 'CAPTURE_REQUESTED'

        # after some time final status is captured
        diagnostic_response = payment.diagnostic(bank_data=response.bank_data)
        assert diagnostic_response['payment_details']['status'] == 'CAPTURED'

    def test_authorization_cancel(self, payment, rsps_recorder):
        trans_id, _, _ = payment.request(
            amount='10.10', orderid='#1234', transaction_id='abcd5678', authorization_mode='PRE_AUTHORIZATION'
        )
        if len(rsps_recorder.calls):
            last_call = json.loads(rsps_recorder.calls[-1].request.body)
            assert last_call == {
                'cardPaymentMethodSpecificInput': {
                    'authorizationMode': 'PRE_AUTHORIZATION',
                },
                'hostedCheckoutSpecificInput': {
                    'returnUrl': 'https://example.com/return',
                    'showResultPage': True,
                    'allowedNumberOfPaymentAttempts': 1,
                    'sessionTimeout': 300,
                },
                'order': {
                    'amountOfMoney': {'amount': 1010, 'currencyCode': 'EUR'},
                    'references': {'merchantReference': 'abcd5678 #1234'},
                },
            }

        # wait for payment...
        query_string = f'RETURNMAC=dd70af3f-31ed-4366-a5c9-d8e2f8e0fa55&hostedCheckoutId={trans_id}'
        response = payment.response(query_string, redirect=True, order_id_hint='#1234')
        assert response.test
        assert response.order_id == trans_id
        assert response.transaction_id == trans_id
        assert (
            response.bank_data['hosted_checkout']['createdPaymentOutput']['payment']['status']
            == 'PENDING_CAPTURE'
        )
        assert response.result == common.PAID

        # cancel the authorization, cancelation can be full or partial.
        cancel_response = payment.cancel('10.10', response.bank_data)
        assert cancel_response['cancel']['payment']['status'] == 'CANCELLED'
        assert (
            cancel_response['hosted_checkout']['createdPaymentOutput']['payment']['status']
            == 'PENDING_CAPTURE'
        )
        assert cancel_response['payment_details']['status'] == 'CANCELLED'
        diagnostic_response = payment.diagnostic(bank_data=response.bank_data)
        assert diagnostic_response['payment_details']['status'] == 'CANCELLED'

    def test_payment_status(self, payment, rsps_recorder):
        response = payment.payment_status('4315597345')
        assert response.test
        assert response.order_id == '4315597345'
        assert response.transaction_id == '4315597345'
        assert response.bank_status == 'CAPTURED'
        assert (
            response.bank_data['hosted_checkout']['createdPaymentOutput']['payment']['status'] == 'CAPTURED'
        )
        assert response.result == common.PAID

    def test_payment_status_before_payment(self, payment, rsps_recorder):
        trans_id, _, _ = payment.request(
            amount='10.10',
            orderid='#1234',
            transaction_id='abcd5678',
            email='john.doe@example.com',
            first_name='John',
            last_name='Doe',
            phone='+33641414141',
        )
        response = payment.payment_status('4315597345')
        assert response.result == common.RECEIVED

    def test_direct_sale_cancelled_by_user(self, payment, rsps_recorder):
        trans_id, _, _ = payment.request(
            amount='10.10',
            orderid='#1234',
            transaction_id='abcd5678',
        )

        # wait for payment...
        query_string = f'RETURNMAC=dd70af3f-31ed-4366-a5c9-d8e2f8e0fa55&hostedCheckoutId={trans_id}'
        response = payment.response(query_string, redirect=True, order_id_hint='#1234')
        assert response.test
        assert response.order_id == trans_id
        assert response.transaction_id == trans_id
        assert response.bank_status == 'CANCELLED'
        assert (
            response.bank_data['hosted_checkout']['createdPaymentOutput']['payment']['status'] == 'CANCELLED'
        )
        assert response.result == common.CANCELLED

    def test_guess(self, payment):
        query_string = 'RETURNMAC=dd70af3f-31ed-4366-a5c9-d8e2f8e0fa55&hostedCheckoutId=1234'
        assert payment.guess(query_string=query_string) == ('worldline', '1234')



# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import html
import logging
import os
import os.path
import random
import sys
from datetime import date
from decimal import ROUND_DOWN, Decimal
from gettext import gettext as _

try:
    if 'django' in sys.modules:
        from django.utils.translation import gettext_lazy as _
except ImportError:
    pass


__all__ = [
    'PaymentCommon',
    'URL',
    'HTML',
    'RANDOM',
    'RECEIVED',
    'ACCEPTED',
    'PAID',
    'ERROR',
    'WAITING',
    'STATUS_DESCRIPTION',
    'STATUS_NAME',
]


RANDOM = random.SystemRandom()

URL = 1
HTML = 2
FORM = 3

RECEIVED = 1
ACCEPTED = 2
PAID = 3
DENIED = 4
CANCELLED = 5
CANCELED = 5  # typo for backward compatibility
WAITING = 6
EXPIRED = 7
ERROR = 99

STATUS_DESCRIPTION = {
    RECEIVED: _('started'),
    ACCEPTED: _('authorized'),
    PAID: _('captured'),
    DENIED: _('refused'),
    CANCELLED: _('cancelled'),
    WAITING: _('running'),
    EXPIRED: _('expired'),
    ERROR: _('error'),
}

STATUS_NAME = {
    RECEIVED: 'RECEIVED',
    ACCEPTED: 'ACCEPTED',
    PAID: 'PAID',
    DENIED: 'DENIED',
    CANCELLED: 'CANCELLED',
    WAITING: 'WAITING',
    EXPIRED: 'EXPIRED',
    ERROR: 'ERROR',
}


# separator between order and transaction ids
ORDERID_TRANSACTION_SEPARATOR = '!'


def force_text(s, encoding='utf-8'):
    if issubclass(type(s), str):
        return s
    try:
        if not issubclass(type(s), str):
            if isinstance(s, bytes):
                s = str(s, encoding)
            else:
                s = str(s)
        else:
            s = s.decode(encoding)
    except UnicodeDecodeError:
        return str(s, encoding, 'ignore')
    return s


def force_byte(s, encoding='utf-8'):
    if isinstance(s, bytes):
        return s
    try:
        return s.encode(encoding)
    except UnicodeEncodeError:
        return s.encode()


class PaymentException(Exception):
    pass


class ResponseError(PaymentException):
    pass


class BackendNotFound(PaymentException):
    pass


class PaymentResponse:
    """Holds a generic view on the result of payment transaction response.

    result -- holds the declarative result of the transaction, does not use
    it to validate the payment in your backoffice, it's just for informing
    the user that all is well.
    test -- indicates if the transaction was a test
    signed -- holds whether the message was signed
    bank_data -- a dictionnary containing some data depending on the bank,
    you have to log it for audit purpose.
    return_content -- when handling a response in a callback endpoint, i.e.
    a response transmitted directly from the bank to the merchant website,
    you usually have to confirm good reception of the message by returning a
    properly formatted response, this is it.
    bank_status -- if result is False, it contains the reason
    order_id -- the id given by the merchant in the payment request
    transaction_id -- the id assigned by the bank to this transaction, it
    could be the one sent by the merchant in the request, but it is usually
    an identifier internal to the bank.
    """

    def __init__(
        self,
        result=None,
        signed=None,
        bank_data=dict(),
        return_content=None,
        bank_status='',
        transaction_id='',
        order_id='',
        test=False,
        transaction_date=None,
    ):
        self.result = result
        self.signed = signed
        self.bank_data = bank_data
        self.return_content = return_content
        self.bank_status = bank_status
        self.transaction_id = transaction_id
        self.order_id = order_id
        self.test = test
        self.transaction_date = transaction_date

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, repr(self.__dict__))

    def is_received(self):
        return self.result == RECEIVED

    def is_accepted(self):
        return self.result == ACCEPTED

    def is_paid(self):
        return self.result == PAID

    def is_error(self):
        return self.result == ERROR

    def as_dict(self):
        serialization = self.__dict__.copy()
        # serialize datetimes
        if hasattr(self.transaction_date, 'isoformat'):
            serialization['transaction_date'] = self.transaction_date.isoformat()
        return serialization


class PaymentCommon:
    has_free_transaction_id = False
    has_empty_response = False

    PATH = '/tmp'
    BANK_ID = '__bank_id'

    def __init__(self, options, logger=None):
        self.logger = logger or logging.getLogger(self.__class__.__module__)
        self.logger.debug('initializing with options %s', options)
        for value in self.description['parameters']:
            key = value['name']
            if 'default' in value:
                setattr(self, key, options.get(key, None) or value['default'])
            else:
                setattr(self, key, options.get(key))

    def transaction_id(self, length, choices, *prefixes):
        while True:
            parts = [RANDOM.choice(choices) for x in range(length)]
            id = ''.join(parts)
            name = '%s_%s_%s' % (str(date.today()), '-'.join(prefixes), str(id))
            try:
                fd = os.open(os.path.join(self.PATH, name), os.O_CREAT | os.O_EXCL)
            except Exception:
                raise
            else:
                os.close(fd)
                return id

    @staticmethod
    def clean_amount(amount, min_amount=0, max_amount=None, cents=True):
        try:
            amount = Decimal(amount)
        except ValueError:
            raise ValueError(
                'invalid amount %s: it must be a decimal integer with two digits '
                'at most after the decimal point',
                amount,
            )
        if int(amount) < min_amount or (max_amount and int(amount) > max_amount):
            raise ValueError('amount %s is not in range [%s, %s]' % (amount, min_amount, max_amount))
        if cents:
            amount *= Decimal('100')  # convert to cents
            amount = amount.to_integral_value(ROUND_DOWN)
            amount = '%03d' % amount
        return str(amount)


class Form:
    def __init__(self, url, method, fields, encoding='utf-8', submit_name='Submit', submit_value='Submit'):
        self.url = url
        self.method = method
        self.fields = fields
        self.encoding = encoding
        self.submit_name = submit_name
        self.submit_value = submit_value

    def __repr__(self):
        s = '%s(' % self.__class__.__name__
        kwargs = []
        for k in ('url', 'method', 'fields', 'encoding', 'submit_name', 'submit_value'):
            kwargs.append('%s=%r' % (k, getattr(self, k)))
        s += ', '.join(kwargs)
        s += ')'
        return s

    def escape(self, s):
        return html.escape(force_text(s, self.encoding))

    def __str__(self):
        s = '<form method="%s" action="%s">' % (self.method, self.url)
        for field in self.fields:
            s += '\n  <input type="%s" name="%s" value="%s"/>' % (
                self.escape(field['type']),
                self.escape(field['name']),
                self.escape(field['value']),
            )
        s += '\n  <input type="submit"'
        if self.submit_name:
            s += ' name="%s"' % self.escape(self.submit_name)
        if self.submit_value:
            s += ' value="%s"' % self.escape(self.submit_value)
        s += ' />'
        s += '\n</form>'
        return s



#!/usr/bin/env python

'''
Setup script for eopayment
'''

import doctest
import io
import os
import subprocess
import sys
from glob import glob
from os.path import basename
from os.path import join as pjoin
from os.path import splitext
from unittest import TestLoader, TextTestRunner

from setuptools import Command, setup
from setuptools.command.build import build as _build
from setuptools.command.install_lib import install_lib as _install_lib
from setuptools.command.sdist import sdist


class TestCommand(Command):
    user_options = []

    def initialize_options(self):
        self._dir = os.getcwd()

    def finalize_options(self):
        pass

    def run(self):
        """
        Finds all the tests modules in tests/, and runs them.
        """
        testfiles = []
        for t in glob(pjoin(self._dir, 'tests', '*.py')):
            if not t.endswith('__init__.py'):
                testfiles.append('.'.join(['tests', splitext(basename(t))[0]]))

        tests = TestLoader().loadTestsFromNames(testfiles)
        import eopayment

        tests.addTests(doctest.DocTestSuite(eopayment))
        t = TextTestRunner(verbosity=4)
        t.run(tests)


class eo_sdist(sdist):
    def run(self):
        print('creating VERSION file')
        if os.path.exists('VERSION'):
            os.remove('VERSION')
        version = get_version()
        version_file = open('VERSION', 'w')
        version_file.write(version)
        version_file.close()
        sdist.run(self)
        print('removing VERSION file')
        if os.path.exists('VERSION'):
            os.remove('VERSION')


def get_version():
    """Use the VERSION, if absent generates a version with git describe, if not
    tag exists, take 0.0.0- and add the length of the commit log.
    """
    if os.path.exists('VERSION'):
        with open('VERSION') as v:
            return v.read()
    if os.path.exists('.git'):
        p = subprocess.Popen(
            ['git', 'describe', '--dirty', '--match=v*'], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        result = p.communicate()[0]
        if p.returncode == 0:
            result = result.decode('ascii').strip()[1:]  # strip spaces/newlines and initial v
            if '-' in result:  # not a tagged version
                try:
                    real_number, commit_count, commit_hash = result.split('-', 2)
                except ValueError:
                    real_number, commit_hash = result.split('-', 2)
                    commit_count = 0
                version = '%s.post%s+%s' % (real_number, commit_count, commit_hash)
            else:
                version = result.replace('.dirty', '+dirty')
            return version
        else:
            return '0.0.post%s' % len(subprocess.check_output(['git', 'rev-list', 'HEAD']).splitlines())

    return '0.0.0'


class compile_translations(Command):
    description = 'compile message catalogs to MO files via django compilemessages'
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        curdir = os.getcwd()
        try:
            from django.core.management import call_command

            for path, dirs, files in os.walk('eopayment'):
                if 'locale' not in dirs:
                    continue
                os.chdir(os.path.realpath(path))
                call_command('compilemessages')
        except ImportError:
            sys.stderr.write('!!! Please install Django >= 3.2 to build translations\n')
        finally:
            os.chdir(curdir)


class build(_build):
    sub_commands = [('compile_translations', None)] + _build.sub_commands


class install_lib(_install_lib):
    def run(self):
        self.run_command('compile_translations')
        _install_lib.run(self)


setup(
    name='eopayment',
    version=get_version(),
    license='GPLv3 or later',
    description='Common API to use all French online payment credit card ' 'processing services',
    include_package_data=True,
    long_description=open(os.path.join(os.path.dirname(__file__), 'README.txt'), encoding='utf-8').read(),
    long_description_content_type='text/plain',
    url='http://dev.entrouvert.org/projects/eopayment/',
    author="Entr'ouvert",
    author_email='info@entrouvert.com',
    maintainer='Benjamin Dauvergne',
    maintainer_email='bdauvergne@entrouvert.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
    ],
    packages=['eopayment'],
    install_requires=[
        'pycryptodomex',
        'requests',
        'click',
        'zeep >= 2.5',
    ],
    cmdclass={
        'build': build,
        'compile_translations': compile_translations,
        'install_lib': install_lib,
        'sdist': eo_sdist,
    },
)



# eopayment - online payment library
# Copyright (C) 2011-2020 Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import base64
import datetime
import hashlib
import hmac
import uuid
from urllib.parse import parse_qs

import requests

from .common import (
    CANCELLED,
    DENIED,
    EXPIRED,
    PAID,
    RECEIVED,
    URL,
    WAITING,
    PaymentCommon,
    PaymentException,
    PaymentResponse,
    _,
)

__all__ = ['Payment']


DEFAULT_TRANSACTION_DURATION = 300  # 5 minutes


class WorldlineAPI:
    def __init__(self, *, url: str, pspid: str, api_key: str, api_secret: str):
        self.url = url
        self.pspid = pspid
        self.api_key = api_key
        self.api_secret = api_secret

    def _worldline_make_request(
        self, endpoint: str, payload: dict = None, method='POST', idempotency_key: str = None
    ):
        """Make a request to Worldline API at the specified endpoint.

        Note: self.ensure_one()

        :param str endpoint: The endpoint to be reached by the request.
        :param dict payload: The payload of the request.
        :param str method: The HTTP method of the request.
        :param str idempotency_key: The idempotency key to pass in the request.
        :return: The JSON-formatted content of the response.
        :rtype: dict
        :raise ValidationError: If an HTTP error occurs.
        """

        pspid = self.pspid
        path = f'/v2/{pspid}/{endpoint}'
        url = f'{self.url}{path}'
        content_type = 'application/json; charset=utf-8' if method == 'POST' else ''
        dt = datetime.datetime.now(datetime.timezone.utc).strftime(
            '%a, %d %b %Y %H:%M:%S GMT'
        )  # Datetime in RFC1123.
        signature = self._worldline_calculate_signature(
            method, endpoint, content_type, dt, idempotency_key=idempotency_key
        )
        api_key = self.api_key
        authorization_header = f'GCS v1HMAC:{api_key}:{signature}'
        headers = {
            'Authorization': authorization_header,
            'Date': dt,
            'Content-Type': content_type,
        }
        if method == 'POST' and idempotency_key:
            headers['X-GCS-Idempotence-Key'] = idempotency_key
        try:
            response = requests.request(method, url, json=payload, headers=headers, timeout=10)
            try:
                response.raise_for_status()
            except requests.exceptions.HTTPError:
                msg = ', '.join([error.get('message', '') for error in response.json().get('errors', [])])
                raise PaymentException(
                    'Worldline: ' + _('The communication with the API failed. Details: %s') % msg
                )
        except (requests.exceptions.ConnectionError, requests.exceptions.Timeout):
            raise PaymentException('Worldline: ' + _('Could not establish the connection to the API.'))
        return response.json()

    def _worldline_calculate_signature(
        self, method: str, endpoint: str, content_type: str, dt_rfc: str, idempotency_key: str = None
    ):
        """Compute the signature for the provided data.

        See https://docs.direct.worldline-solutions.com/en/integration/api-developer-guide/authentication.

        :param str method: The HTTP method of the request
        :param str endpoint: The endpoint to be reached by the request.
        :param str content_type: The 'Content-Type' header of the request.
        :param datetime.datetime dt_rfc: The timestamp of the request, in RFC1123 format.
        :param str idempotency_key: The idempotency key to pass in the request.
        :return: The calculated signature.
        :rtype: str
        """
        # specific order required: method, content_type, date, custom headers, endpoint
        values_to_sign = [method, content_type, dt_rfc]
        if idempotency_key:
            values_to_sign.append(f'x-gcs-idempotence-key:{idempotency_key}')
        values_to_sign.append(f'/v2/{self.pspid}/{endpoint}')

        signing_str = '\n'.join(values_to_sign) + '\n'
        api_secret = self.api_secret
        signature = hmac.new(api_secret.encode(), signing_str.encode(), hashlib.sha256)
        return base64.b64encode(signature.digest()).decode()

    def hosted_checkout(
        self,
        *,
        amount: int,
        return_url: str,
        reference: str,
        first_name='',
        last_name='',
        email='',
        phone='',
        currency_code='EUR',
        show_result_page=True,
        webhook_url='',
        authorization_mode='SALE',
        session_timeout=DEFAULT_TRANSACTION_DURATION,  # 5 minutes to pay
        allowed_number_of_payment_attempts=1,
    ):
        # https://docs.direct.worldline-solutions.com/en/api-reference#tag/HostedCheckout/operation/CreateHostedCheckoutApi
        assert isinstance(amount, int)
        payload = {
            'hostedCheckoutSpecificInput': {
                'returnUrl': return_url,
                'showResultPage': show_result_page,
                'sessionTimeout': session_timeout,
                'allowedNumberOfPaymentAttempts': allowed_number_of_payment_attempts,
            },
            'order': {
                'amountOfMoney': {
                    # en centimes
                    'amount': amount,
                    'currencyCode': currency_code,
                },
                'references': {
                    'merchantReference': reference,
                },
            },
        }
        if email:
            payload['order'].setdefault('customer', {}).setdefault('contactDetails', {})[
                'emailAddress'
            ] = email
        if phone:
            payload['order'].setdefault('customer', {}).setdefault('contactDetails', {})[
                'phoneNumber'
            ] = phone
        if first_name:
            payload['order'].setdefault('customer', {}).setdefault('personalInformation', {}).setdefault(
                'name', {}
            )['firstName'] = first_name
        if last_name:
            payload['order'].setdefault('customer', {}).setdefault('personalInformation', {}).setdefault(
                'name', {}
            )['surName'] = last_name
        if webhook_url:
            payload.setdefault('feedbacks', {})['webhookUrl'] = webhook_url
        if authorization_mode:
            assert authorization_mode in ['PRE_AUTHORIZATION', 'SALE']
            payload.setdefault('cardPaymentMethodSpecificInput', {})['authorizationMode'] = authorization_mode
        return self._worldline_make_request('hostedcheckouts', payload=payload)

    def hosted_checkout_status(self, *, hosted_checkout_id):
        # https://docs.direct.worldline-solutions.com/en/api-reference#tag/HostedCheckout/operation/GetHostedCheckoutApi
        return self._worldline_make_request(f'hostedcheckouts/{hosted_checkout_id}', method='GET')

    def payment_capture(self, amount, payment_id, idempotency_key=None):
        # https://docs.direct.worldline-solutions.com/en/api-reference#tag/Payments/operation/CapturePaymentApi

        payload = {
            'amount': amount,
        }
        return self._worldline_make_request(
            f'payments/{payment_id}/capture', payload=payload, idempotency_key=idempotency_key
        )

    def payment_cancel(self, amount, payment_id, idempotency_key=None, currency_code='EUR'):
        # https://docs.direct.worldline-solutions.com/en/api-reference#tag/Payments/operation/CancelPaymentApi
        payload = {
            'amountOfMoney': {
                'amount': amount,
                'currencyCode': currency_code,
            },
            'isFinal': True,
        }
        return self._worldline_make_request(
            f'payments/{payment_id}/cancel', payload=payload, idempotency_key=idempotency_key
        )

    def payment_details(self, payment_id):
        return self._worldline_make_request(f'payments/{payment_id}/details', method='GET')


class Payment(PaymentCommon):
    '''Implements Worldline API, see .https://docs.direct.worldline-solutions.com/en/integration/basic-integration-methods/hosted-checkout-page'''

    preprod_url = 'https://payment.preprod.direct.worldline-solutions.com'
    prod_url = 'https://payment.direct.worldline-solutions.com'

    min_time_between_transactions = DEFAULT_TRANSACTION_DURATION  # 20 minutes

    description = {
        'caption': 'Worldline API v2',
        'parameters': [
            {
                'name': 'environment',
                'caption': _('Environment'),
                'type': str,
                'choices': ['preprod', 'prod'],
            },
            {
                'name': 'pspid',
                'caption': _('PSPID'),
                'type': str,
                'required': True,
            },
            {
                'name': 'api_key',
                'caption': _('API key'),
                'required': True,
                'type': str,
            },
            {
                'name': 'api_secret',
                'caption': _('API secret'),
                'required': True,
            },
            {
                'name': 'normal_return_url',
                'caption': _('Normal return URL'),
                'default': '',
                'required': True,
            },
            {
                'name': 'automatic_return_url',
                'caption': _('Automatic return URL'),
                'required': False,
            },
        ],
    }

    def get_api(self):
        if self.environment == 'prod':
            url = self.prod_url
        elif self.environment == 'preprod':
            url = self.preprod_url
        else:
            raise NotImplementedError
        return WorldlineAPI(url=url, pspid=self.pspid, api_key=self.api_key, api_secret=self.api_secret)

    def request(
        self,
        amount,
        name=None,
        first_name=None,
        last_name=None,
        address=None,
        email=None,
        phone=None,
        orderid=None,
        info1=None,
        info2=None,
        info3=None,
        next_url=None,
        transaction_id=None,
        authorization_mode='SALE',
        session_timeout=DEFAULT_TRANSACTION_DURATION,  # 5 minutes to pay
        allowed_number_of_payment_attempts=1,
        **kwargs,
    ):
        amount = int(self.clean_amount(amount))
        api = self.get_api()
        transaction_id = transaction_id or uuid.uuid4().hex
        hosted_checkout_response = api.hosted_checkout(
            amount=amount,
            return_url=self.normal_return_url,
            reference=f'{transaction_id} {orderid or ""}',
            first_name=first_name,
            last_name=last_name,
            email=email,
            phone=phone,
            authorization_mode=authorization_mode,
            session_timeout=session_timeout,
            allowed_number_of_payment_attempts=allowed_number_of_payment_attempts,
        )
        return hosted_checkout_response['hostedCheckoutId'], URL, hosted_checkout_response['redirectUrl']

    def response(self, query_string, redirect=False, order_id_hint=None, order_status_hint=None, **kwargs):
        # ex.: ?RETURNMAC=dd70af3f-31ed-4366-a5c9-d8e2f8e0fa55&hostedCheckoutId=4311628069
        fields = parse_qs(query_string, True)
        hosted_checkout_id = (fields.get('hostedCheckoutId') or [order_id_hint])[0]
        return self.get_hosted_checkout_response(hosted_checkout_id)

    def get_hosted_checkout_response(self, hosted_checkout_id):
        api = self.get_api()
        hosted_checkout_details = api.hosted_checkout_status(hosted_checkout_id=hosted_checkout_id)
        status = hosted_checkout_details['status']
        created_payment_output = hosted_checkout_details['createdPaymentOutput']
        if not created_payment_output:
            return PaymentResponse(
                result=RECEIVED,
                signed=True,
                bank_data={},
                order_id=hosted_checkout_id,
                bank_status=status,
                test=self.environment == 'preprod',
            )

        payment_id = created_payment_output['payment']['id']
        payment_details = api.payment_details(payment_id=payment_id)

        # https://docs.direct.worldline-solutions.com/en/integration/api-developer-guide/statuses
        payment_status = payment_details['status']
        if status == 'PAYMENT_CREATED':
            status_category = hosted_checkout_details['createdPaymentOutput']['paymentStatusCategory']
            if status_category == 'SUCCESSFUL':
                eopayment_result = PAID
            elif status_category == 'STATUS_UNKNOWN':
                eopayment_result = RECEIVED
            elif status_category == 'REJECTED':
                if payment_status == 'CREATED':
                    eopayment_result = WAITING
                elif payment_status == 'CANCELLED':
                    eopayment_result = CANCELLED
                elif payment_status in ('REJECTED', 'REJECTED_CAPTURE'):
                    eopayment_result = DENIED
                else:
                    raise NotImplementedError
        elif status == 'IN_PROGRESS':
            eopayment_result = WAITING
        elif status == 'CANCELLED_BY_CONSUMER':
            eopayment_result = CANCELLED
        else:
            raise NotImplementedError
        return PaymentResponse(
            result=eopayment_result,
            signed=True,
            bank_data={
                'hosted_checkout': hosted_checkout_details,
                'payment_details': payment_details,
            },
            order_id=hosted_checkout_id,
            transaction_id=hosted_checkout_id,
            bank_status=payment_status,
            test=self.environment == 'preprod',
        )

    def _get_payment_id(self, bank_data):
        payment_id = (
            bank_data.get('hosted_checkout', {}).get('createdPaymentOutput', {}).get('payment', {}).get('id')
        )
        if not payment_id:
            raise ResponseError(f'missing payment_id in bank_data {bank_data}')
        return payment_id

    def cancel(self, amount, bank_data, **kwargs):
        amount = int(self.clean_amount(amount))
        payment_id = self._get_payment_id(bank_data)

        # Check that the payment is currently waiting for the capture
        payment_details = self.get_api().payment_details(payment_id=payment_id)
        status = payment_details['status']
        if not status == 'PENDING_CAPTURE':
            return {**bank_data, 'payment_details': payment_details}

        cancel_response = self.get_api().payment_cancel(amount, payment_id=payment_id)
        payment_details = self.get_api().payment_details(payment_id=payment_id)
        return {**bank_data, 'cancel': cancel_response, 'payment_details': payment_details}

    def validate(self, amount, bank_data, **kwargs):
        amount = int(self.clean_amount(amount))
        payment_id = self._get_payment_id(bank_data)

        # Check that the payment is currently waiting for the capture
        payment_details = self.get_api().payment_details(payment_id=payment_id)
        status = payment_details['status']
        if not status == 'PENDING_CAPTURE':
            return {**bank_data, 'payment_details': payment_details} | bank_data

        capture_response = self.get_api().payment_capture(amount, payment_id=payment_id)
        payment_details = self.get_api().payment_details(payment_id=payment_id)
        return {**bank_data, 'capture': capture_response, 'payment_details': payment_details}

    def payment_status(self, transaction_id, transaction_date=None, **kwargs):
        if transaction_date:
            if transaction_date.tzinfo:  # date is aware
                now = datetime.datetime.now(tz=datetime.datetime.utc)
            else:
                now = datetime.datetime.now()
            delta = now - transaction_date
        else:
            delta = datetime.timedelta(seconds=0)
        threshold = datetime.timedelta(seconds=DEFAULT_TRANSACTION_DURATION)

        response = self.get_hosted_checkout_response(transaction_id)
        # if transaction has been started for more thant 5 minutes, consider it
        # expired
        if response.result in [WAITING, RECEIVED] and delta > threshold:
            response.result = EXPIRED
        return response

    def diagnostic(self, *, bank_data, **kwargs):
        payment_id = (
            bank_data.get('hosted_checkout', {}).get('createdPaymentOutput', {}).get('payment', {}).get('id')
        )
        if not payment_id:
            raise ResponseError(f'missing payment_id in bank_data {bank_data}')
        api = self.get_api()
        payment_details = api.payment_details(payment_id=payment_id)
        return {**bank_data, 'payment_details': payment_details}

    @classmethod
    def guess(cls, *, method=None, query_string=None, body=None, headers=None, backends=(), **kwargs):
        for content in [query_string, body]:
            if isinstance(content, bytes):
                try:
                    content = content.decode()
                except UnicodeDecodeError:
                    pass
            if isinstance(content, str):
                fields = parse_qs(content)
                if 'hostedCheckoutId' in fields:
                    return fields['hostedCheckoutId'][0]
        return None



from pathlib import Path

import nox


def run_hook(name, *args, **kwargs):
    for file in [Path(__name__).parent / '.nox-hooks.py', Path('~/.config/nox/eo-hooks.py').expanduser()]:
        if not file.exists():
            continue

        globals_ = {}
        exec(file.read_text(), globals_)
        hook = globals_.get(name, None)
        if hook:
            hook(*args, **kwargs)


def setup_venv(session, *packages):
    packages = [
        'coverage',
        'pytest',
        'pytest-cov',
        'pytest-freezer',
        'mock<4',
        'httmock',
        'lxml',
        'responses',
        *packages,
    ]
    run_hook('setup_venv', session, packages)
    session.install('-e', '.', *packages, silent=False)


def hookable_run(session, *args, **kwargs):
    args = list(args)
    run_hook('run', session, args, kwargs)
    session.run(*args, **kwargs)


@nox.session()
def tests(session):
    setup_venv(session)

    args = ['py.test']
    if '--coverage' in session.posargs or not session.interactive:
        while '--coverage' in session.posargs:
            session.posargs.remove('--coverage')
        args += [
            '--cov=eopayment/',
            '--cov-report=term-missing',
        ]
        if not session.interactive:
            args += [
                '-v',
                '--cov-report=xml',
                '--cov-report=html',
                '--junitxml=junit-coverage.xml',
            ]

    args += session.posargs + ['tests/']

    hookable_run(
        session,
        *args,
    )


@nox.session
def codestyle(session):
    session.install('pre-commit')
    session.run('pre-commit', 'run', '--all-files', '--show-diff-on-failure')


@nox.session
def check_manifest(session):
    # django is only required to compile messages
    session.install('django', 'check-manifest')
    # compile messages and css
    ignores = [
        'VERSION',
    ]
    session.run('check-manifest', '--ignore', ','.join(ignores))
